# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"

import numpy as np
import re
import datetime
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.colors as mpl_colors
import matplotlib.ticker as mpl_ticker
import matplotlib.dates as mpl_dates
from scipy.interpolate import interp1d

from geospacelab.visualization.mpl.__base__ import PanelBase
from geospacelab.datahub.__variable_base__ import VariableBase as VariableModel
import geospacelab.toolbox.utilities.numpyarray as arraytool
import geospacelab.visualization.mpl.axis_ticks as ticktool
import geospacelab.toolbox.utilities.pybasic as basic
import geospacelab.toolbox.utilities.pylogging as mylog
from geospacelab.visualization.mpl._helpers import check_panel_ax
import geospacelab.visualization.mpl.colormaps as mycmap


class Panel(PanelBase):

    def __init__(self, *args, figure=None, from_subplot=True, **kwargs):
        super(Panel, self).__init__(*args, figure=figure, from_subplot=from_subplot, **kwargs)


class TSPanel(Panel):

    _default_plot_config = {
            'linestyle': '-',
            'linewidth': 1.5,
            'marker': '.',
            'markersize': 1
        }

    _default_legend_config = {
        'loc': 'upper left',
        'bbox_to_anchor': (1.0, 1.0),
        'frameon': False,
        'fontsize': 'small'
    }
    _default_xtick_params = {
        'labelsize': plt.rcParams['xtick.labelsize'],
        'direction': 'inout',
    }
    _default_colorbar_offset = 0.1

    def __init__(
            self, *args, 
            dt_fr=None, dt_to=None, figure=None, from_subplot=True,
            bottom_panel=True, timeline_reverse=False, timeline_extra_labels=None,
            time_gap=True,
            **kwargs
    ):
        
        super(Panel, self).__init__(*args, figure=figure, from_subplot=from_subplot, **kwargs)

        self.dt_fr = dt_fr
        self.dt_to = dt_to

        self._xlim = [dt_fr, dt_to]
        self.timeline_reverse = timeline_reverse
        self.time_gap = time_gap
        if timeline_extra_labels is None:
            timeline_extra_labels = []
        self.timeline_extra_labels = timeline_extra_labels
        self._var_for_config = None
        self.bottom_panel = bottom_panel

    def draw(self, plot_layout, ax=None):
        self.overlay_from_variables(plot_layout, ax=ax)

    def overlay_from_variables(self, layout_in, ax=None, level=0):
        if ax is None:
            ax = self()
        self.sca(ax)

        if level == 0 and not list(layout_in):
            raise TypeError
        elif level == 0:
            if type(layout_in[0]) is list:
                raise ValueError("The first element in the plot layout cannot be a list!")
            if any(type(i) is list for i in layout_in):
                self.axes_overview[ax]['twinx'] = 'on'

        if level > 1:
            raise NotImplemented

        iplts = []

        for ind, elem in enumerate(layout_in):

            if ind == 0 or issubclass(elem.__class__, VariableModel):
                var = elem[0] if type(elem) is list else elem
                iplts_add = self.overlay_a_variable(var, ax=ax)
            elif level == 0 and isinstance(elem, list):
                ax_in = self.add_twin_axes(
                    ax=ax, which='x', location='right', offset_type='outward',
                    offset=50*(len(self.axes_overview[ax]['twinx_axes']))
                )
                self.axes_overview[ax_in]['twinx'] = 'self'
                ax_in._get_lines.prop_cycler = ax._get_lines.prop_cycler
                iplts_add = self.overlay_from_variables(elem, ax=ax_in, level=level+1)
            else:
                raise NotImplementedError
            iplts.extend(iplts_add)

        if level == 0:
            if self.axes_overview[ax]['legend'] == 'on' and self.axes_overview[ax]['twinx'] == 'off':
                self._check_legend(ax)
            if self.axes_overview[ax]['twinx'] == 'on':
                self._check_twinx(ax)
            if self.axes_overview[ax]['colorbar'] == 'on':
                self._check_colorbar(ax)
            self._set_xaxis(ax=ax)

        self._set_yaxis(ax=ax)
        return iplts

    def overlay_a_variable(self, var, ax=None):
        self.axes_overview[ax]['variables'].extend([var])
        var_for_config = self.axes_overview[ax]['variables'][0]
        plot_style = var_for_config.visual.plot_config.style

        if plot_style in ['1P', '1noE']:
            iplt = self.overlay_line(var, ax=ax)
        elif plot_style in ['1', '1E']:
            iplt = self.overlay_line(var, ax=ax, errorbar='on')
        elif plot_style in ['2P']:
            iplt = self.overlay_pcolormesh(var, ax=ax)
        elif plot_style in ['2I']:
            iplt = self.overlay_imshow(var, ax=ax)
        elif plot_style in ['1B']:
            iplt = self.overlay_bar(var, ax=ax)
        elif plot_style in ['1F0']:
            iplt = self.overlay_fill_between_y_zero(var, ax=ax)
        elif plot_style in ['1S']:
            iplt = self.overlay_scatter(var, ax=ax)
        elif plot_style in ['1C']:
            iplt = self.overlay_multiple_colored_line(var, ax=ax)
        else:
            raise NotImplementedError

        return iplt

    @check_panel_ax
    def overlay_line(self, var, ax=None, errorbar='off', **kwargs):
        """
        Overlay a line plot in the axes.
        :param var: A GeospaceLab Variable object
        :param ax: The axes to plot.
        :param errorbar: If 'on', show errorbar.
        :param kwargs: Other keyword arguments forwarded to ax.plot() or ax.errorbar()
        :return:
        """
        il = None

        data = self._retrieve_data_1d(var)
        x = data['x']
        y = data['y']
        y_err = data['y_err']

        plot_config = basic.dict_set_default(kwargs, **var.visual.plot_config.line)
        plot_config = basic.dict_set_default(plot_config, **self._default_plot_config)
        if self.axes_overview[ax]['twinx'] in ['on', 'self']:
            var.visual.axis[1].label = '@v.label'
        label = var.get_visual_axis_attr(axis=2, attr_name='label')
        plot_config = basic.dict_set_default(plot_config, label=label)
        if errorbar == 'off':
            il = ax.plot(x, y, **kwargs)
        elif errorbar == 'on':
            errorbar_config = dict(plot_config)
            errorbar_config.update(var.visual.plot_config.errorbar)
            il = ax.errorbar(x, y, yerr=y_err, ax=ax, **errorbar_config)
        if type(il) is not list:
            il = [il]
        self.axes_overview[ax]['lines'].extend(il)
        if self.axes_overview[ax]['legend'] is None:
            self.axes_overview[ax]['legend'] = 'on'
        return il

    @check_panel_ax
    def overlay_bar(self, *args, ax=None, **kwargs):

        var = args[0]
        il = None
        data = self._retrieve_data_1d(var)
        x = data['x'].flatten()
        height = data['y'].flatten()

        bar_config = basic.dict_set_default(kwargs, **var.visual.plot_config.bar)

        color_by_value = bar_config.pop('color_by_value', False)
        colormap = bar_config.pop('colormap', 'viridis')
        vmin = bar_config.pop('vmin', np.nanmin(height))
        vmax = bar_config.pop('vmax', np.nanmax(height))
        if color_by_value:
            if isinstance(colormap, str):
                colormap = mpl.cm.get_cmap(colormap)
            norm = mpl_colors.Normalize(vmin=vmin, vmax=vmax)
            colors = colormap(norm(height))
            bar_config.update(color=colors)

        # plot_config = basic.dict_set_default(plot_config, **self._default_plot_config)
        ib = ax.bar(x, height, **bar_config)
        return [ib]

    @check_panel_ax
    def overlay_pcolormesh(self, *args, ax=None, **kwargs):
        var = args[0]

        data = self._retrieve_data_2d(var)
        x = data['x']
        y = data['y']
        z = data['z']

        if x.shape[0] == z.shape[0]:
            delta_x = np.diff(x, axis=0)
            x[:-1, :] = x[:-1, :] + delta_x/2
            x = np.vstack((
                np.array(x[0, 0] - delta_x[0, 0] / 2).reshape((1, 1)),
                x[:-1, :],
                np.array(x[-1, 0] + delta_x[-1, 0] / 2).reshape((1, 1))
            ))

        if y.shape[1] == z.shape[1]:
            delta_y = np.diff(y, axis=1)
            y[:, :-1] = y[:, :-1] + delta_y/2
            y = np.hstack((
                np.array(y[:, 0] - delta_y[:, 0]/2).reshape((y.shape[0], 1)),
                y[:, :-1],
                np.array(y[:, -1] + delta_y[:, -1]/2).reshape((y.shape[0], 1)),
            ))

        if y.shape[0] == z.shape[0]:
            y = np.vstack((y, y[-1, :].reshape((1, y.shape[1]))))

        pcolormesh_config = var.visual.plot_config.pcolormesh
        z_lim = var.visual.axis[2].lim
        if z_lim is None:
            z_lim = [np.nanmin(z.flatten()), np.nanmax(z.flatten())]
        z_scale = var.visual.axis[2].scale
        if z_scale == 'log':
            norm = mpl_colors.LogNorm(vmin=z_lim[0], vmax=z_lim[1])
            pcolormesh_config.update(norm=norm)
        else:
            pcolormesh_config.update(vmin=z_lim[0])
            pcolormesh_config.update(vmax=z_lim[1])
        colormap = mycmap.get_colormap(var.visual.plot_config.pcolormesh.get('cmap', None))
        pcolormesh_config.update(cmap=colormap)

        im = ax.pcolormesh(x.T, y.T, z.T, **pcolormesh_config)
        self.axes_overview[ax]['collections'].extend([im])
        if self.axes_overview[ax]['colorbar'] is None:
            self.axes_overview[ax]['colorbar'] = 'on'
        return [im]

    @check_panel_ax
    def _get_var_for_config(self, ax=None, ind=0):
        var_for_config = self.axes_overview[ax]['variables'][ind]
        self._var_for_config = var_for_config
        return var_for_config

    def _set_xaxis(self, ax):

        if self._xlim[0] is None:
            self._xlim[0] = mpl_dates.num2date(ax.get_xlim()[0])
        if self._xlim[1] is None:
            self._xlim[1] = mpl_dates.num2date(ax.get_xlim()[1])
        ax.set_xlim(self._xlim)
        # reverse the x axis if timeline_reverse=True
        if self.timeline_reverse:
            ax.set_xlim((self._xlim[1], self._xlim[0]))

        ax.xaxis.set_tick_params(labelsize=self._default_xtick_params['labelsize'])
        ax.xaxis.set_tick_params(
            which='both',
            direction=self._default_xtick_params['direction'],
            bottom=True, top=True
        )
        ax.xaxis.set_tick_params(which='major', length=8)
        ax.xaxis.set_tick_params(which='minor', length=4)

        # use date locators
        # majorlocator, minorlocator, majorformatter = ticktool.set_timeline(self._xlim[0], self._xlim[1])
        from geospacelab.visualization.mpl.axis_ticks import DatetimeMajorFormatter, DatetimeMajorLocator, DatetimeMinorLocator
        majorlocator = DatetimeMajorLocator()
        majorformatter = DatetimeMajorFormatter(majorlocator)

        if not self.bottom_panel:
            majorformatter = mpl_ticker.NullFormatter()
            plt.setp(ax.get_xticklabels(), visible=False)
        ax.xaxis.set_major_locator(majorlocator)
        ax.xaxis.set_major_formatter(majorformatter)
        # minor locator must be set up after setting the major locator
        minorlocator = DatetimeMinorLocator(ax=ax, majorlocator=majorlocator)
        ax.xaxis.set_minor_locator(minorlocator)
        if self.bottom_panel:
            self._set_xaxis_ticklabels(ax, majorformatter=majorformatter)

    def _set_xaxis_ticklabels(self, ax, majorformatter=None):
        var_for_config = self._get_var_for_config(ax=ax)
        # ax.tick_params(axis='x', labelsize=plt.rcParams['xtick.labelsize'])
        # set UT timeline
        if not list(self.timeline_extra_labels):

            ax.set_xlabel('UT', fontsize=12, fontweight='normal')
            return

        figheight = self.figure.get_size_inches()[1]*2.54
        figwidth = self.figure.get_size_inches()[0]*2.54
        xoffset = 0.1 - figwidth * 0.002
        yoffset = 0.07 - figheight * 0.0022

        ticks = ax.get_xticks()
        ylim0, _ = ax.get_ylim()
        xy_fig = []
        # transform from data cs to figure cs
        for tick in ticks:
            px = ax.transData.transform([tick, ylim0])
            xy = self.figure.transFigure.inverted().transform(px)
            xy_fig.append(xy)

        xlabels = ['UT']
        x_depend = var_for_config.get_depend(axis=0, retrieve_data=True)
        x0 = np.array(mpl_dates.date2num(x_depend['UT'])).flatten()
        x1 = np.array(ticks)
        ys = [x1]       # list of tick labels
        for ind, label in enumerate(self.timeline_extra_labels):
            if label in x_depend.keys():
                y0 = x_depend[label].flatten()
            elif label in var_for_config.dataset.keys():
                y0 = var_for_config.dataset[label].value
            else:
                raise KeyError
            if 'MLT' in label.upper():
                y0_sin = np.sin(y0 / 24. * 2 * np.pi)
                y0_cos = np.cos(y0 / 24. * 2 * np.pi)
                itpf_sin = interp1d(x0, y0_sin, bounds_error=False, fill_value='extrapolate')
                itpf_cos = interp1d(x0, y0_cos, bounds_error=False, fill_value="extrapolate")
                y0_sin_i = itpf_sin(x1)
                y0_cos_i = itpf_cos(x1)
                rad = np.sign(y0_sin_i) * (np.pi / 2 - np.arcsin(y0_cos_i))
                rad = np.where((rad >= 0), rad, rad + 2 * np.pi)
                y1 = rad / 2. / np.pi * 24.
            elif 'LON' in label.upper():
                y0_sin = np.sin(y0 * np.pi / 180.)
                y0_cos = np.cos(y0 * np.pi / 180.)
                itpf_sin = interp1d(x0, y0_sin, bounds_error=False, fill_value='extrapolate')
                itpf_cos = interp1d(x0, y0_cos, bounds_error=False, fill_value="extrapolate")
                y0_sin_i = itpf_sin(x1)
                y0_cos_i = itpf_cos(x1)
                rad = np.sign(y0_sin_i) * (np.pi / 2 - np.arcsin(y0_cos_i))
                y1 = rad * 180. / np.pi
            else:
                itpf = interp1d(x0, y0, bounds_error=False, fill_value='extrapolate')
                y1 = itpf(x1)
            ys.append(y1)
            xlabels.append(label)

        for ind, xticks in enumerate(ys):
            ax.text(
                xy_fig[0][0] - xoffset, xy_fig[0][1] - yoffset * ind - 0.013,
                xlabels[ind].replace('_', '/'),
                fontsize=plt.rcParams['xtick.labelsize'], fontweight='normal',
                horizontalalignment='right', verticalalignment='top',
                transform=self.figure.transFigure
            )
            for ind_pos, xtick in enumerate(xticks):
                if np.isnan(xtick):
                    continue
                if ind == 0:
                    text = majorformatter.format_data(xtick)
                elif 'MLT' in xlabels[ind]:
                    text = (datetime.datetime(1970, 1, 1) + datetime.timedelta(hours=xtick)).strftime('%H:%M')
                else:
                    text = '%.1f' % xtick
                ax.text(
                    xy_fig[ind_pos][0], xy_fig[ind_pos][1] - yoffset * ind - 0.013,
                    text,
                    fontsize=plt.rcParams['xtick.labelsize'],
                    horizontalalignment='center', verticalalignment='top',
                    transform=self.figure.transFigure
                )
        ax.xaxis.set_major_formatter(mpl_ticker.NullFormatter())

        # if self.major_timeline == 'MLT':
        #     for ind_pos, xtick in enumerate(ys[mlt_ind]):
        #         if xtick == np.nan:
        #             continue
        #         text = (datetime.datetime(1970, 1, 1) + datetime.timedelta(hours=xtick)).strftime('%H:%M')
        #         plt.text(
        #             xy_fig[ind_pos][0], xy_fig[ind_pos][1] - yoffset * ind - 0.15,
        #             text,
        #             fontsize=9,
        #             horizontalalignment='center', verticalalignment='top',
        #             transform=self.figure.transFigure
        #         )
        #     ax.set_xlabel('MLT')
        #     return



    @check_panel_ax
    def _set_yaxis(self, ax=None):
        var_for_config = self._get_var_for_config(ax=ax)
        ax.tick_params(axis='y', which='major', labelsize=plt.rcParams['ytick.labelsize'])
        # Set y axis lim
        self._set_ylim(ax=ax)

        # set y labels and alignment two methods: fig.align_ylabels(axs[:, 1]) or yaxis.set_label_coords
        ylabel = var_for_config.get_visual_axis_attr('label', axis=1)
        yunit = var_for_config.get_visual_axis_attr('unit', axis=1)
        ylabel_style = var_for_config.get_visual_axis_attr('label_style', axis=1)

        if self.axes_overview[ax]['twinx'] == 'self':
            ylabel = self.generate_label(ylabel, unit=yunit, style='single')
            ax.set_ylabel(ylabel, va='baseline', rotation=270, fontsize=plt.rcParams['axes.labelsize'])
        else:
            ylabel = self.generate_label(ylabel, unit=yunit, style=ylabel_style)
            label_pos = var_for_config.visual.axis[1].label_pos

            ax.set_ylabel(ylabel, va='bottom', fontsize=plt.rcParams['axes.labelsize'])
            if label_pos is not None:
                ax.yaxis.set_label_coords(label_pos[0], label_pos[1])
        ylim = ax.get_ylim()

        # set yaxis scale
        yscale = var_for_config.visual.axis[1].scale
        ax.set_yscale(yscale)

        # set yticks and yticklabels, usually do not change the matplotlib default
        yticks = var_for_config.visual.axis[1].ticks
        yticklabels = var_for_config.visual.axis[1].tick_labels
        if yticks is not None:
            ax.set_yticks(yticks)
            if yticklabels is not None:
                ax.set_yticklabels(yticklabels)
        else:
            if yscale == 'linear':
                major_max = var_for_config.visual.axis[1].major_tick_max
                minor_max = var_for_config.visual.axis[1].minor_tick_max
                if major_max is None:
                    ax.yaxis.set_minor_locator(mpl_ticker.AutoLocator())
                else:
                    ax.yaxis.set_major_locator(mpl_ticker.MaxNLocator(major_max))
                if minor_max is None:
                    ax.yaxis.set_minor_locator(mpl_ticker.AutoMinorLocator())
                else:
                    ax.yaxis.set_minor_locator(mpl_ticker.MaxNLocator(minor_max))

            if yscale == 'log':
                locmin = mpl_ticker.LogLocator(base=10.0,
                                           subs=(0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9),
                                           numticks=12
                                           )
                ax.yaxis.set_minor_locator(locmin)
                # ax.yaxis.set_minor_formatter(mpl.ticker.NullFormatter())
        ax.set_ylim(ylim)

    @check_panel_ax
    def _set_ylim(self, ax=None, zero_line='on'):

        var_for_config = self.axes_overview[ax]['variables'][0]
        ylim = var_for_config.visual.axis[1].lim

        ylim_current = ax.get_ylim()
        if ylim is None:
            ylim = ylim_current
        elif ylim[0] == -np.inf and ylim[1] == np.inf:
            maxlim = np.max(np.abs(ylim_current))
            ylim = [-maxlim, maxlim]
        else:
            if ylim[0] is None:
                ylim[0] = ylim_current[0]
            if ylim[1] is None:
                ylim[1] = ylim_current[1]
        if zero_line == 'on' and self.axes_overview[ax]['twinx']=='off':
            if (ylim[0] < 0) and (ylim[1] > 0):
                ax.plot(ax.get_xlim(), [0, 0], 'k--', linewidth=0.5)
        ax.set_ylim(ylim)
        return

    def _check_legend(self, ax):
        ax_ov = self.axes_overview[ax]
        var_for_config = ax_ov['variables'][0]
        legend_config = var_for_config.visual.plot_config.legend
        # get color
        legend_config = basic.dict_set_default(legend_config, **self._default_legend_config)

        if len(ax_ov['lines']) == 1 and ax_ov['lines'][0]._label == '_line0':
            return

        ax.legend(handles=ax_ov['lines'], **legend_config)

    def _check_twinx(self, ax):
        ax_ov = self.axes_overview[ax]
        axes = [ax]
        axes.extend(ax_ov['twinx_axes'])
        for ind, pax in enumerate(axes):

            pax_ov = self.axes_overview[pax]
            if list(pax_ov['lines']):
                il = pax_ov['lines'][0]
                pax.yaxis.label.set_color(il.get_color())
                pax.tick_params(axis='y', which='both', colors=il.get_color())
                pax.spines['right'].set_edgecolor(il.get_color())
            else:
                continue

    def _check_colorbar(self, ax):
        ax_ov = self.axes_overview[ax]
        var_for_config = ax_ov['variables'][0]
        colorbar_config = var_for_config.visual.plot_config.colorbar

        c_scale = var_for_config.visual.axis[2].scale
        c_label = var_for_config.get_visual_axis_attr('label', axis=2)
        c_unit = var_for_config.get_visual_axis_attr('unit', axis=2)
        c_label_style = var_for_config.get_visual_axis_attr('label_style', axis=2)

        c_label = self.generate_label(c_label, unit=c_unit, style=c_label_style)
        c_ticks = var_for_config.visual.axis[2].ticks
        c_tick_labels = var_for_config.visual.axis[2].tick_labels
        im = ax_ov['collections'][0]

        offset = self._default_colorbar_offset
        ntwinx = len(ax_ov['twinx_axes'])
        cax_position = [1.02 + offset * ntwinx, 0.01, 0.025, 0.85]
        colorbar_config.update(
            cax_scale=c_scale, cax_label=c_label, cax_ticks=c_ticks, cax_tick_labels=c_tick_labels,
            cax_position=cax_position
        )

        self.add_colorbar(im, cax='new', **colorbar_config)

    def _retrieve_data_1d(self, var):
        x_data = var.get_visual_axis_attr(axis=0, attr_name='data')
        if type(x_data) == list:
            x_data = x_data[0]
        if x_data is None:
            depend_0 = var.get_depend(axis=0)
            x_data = depend_0['UT']  # numpy array, type=datetime
            if x_data is None:
                x_data = np.array([0, 1]).reshape(2, 1)
                var.visual.axis[0].mask_gap = False

        y_data = var.get_visual_axis_attr(axis=1, attr_name='data')
        if y_data is None:
            y_data = var.value
            if y_data is None:
                y_data = np.empty_like(x_data, dtype=np.float32)
                y_data[::] = np.nan

        y_err_data = var.get_visual_axis_attr(axis=1, attr_name='data_err')
        if y_err_data is None:
            y_err_data = var.error

        x_data_res = var.visual.axis[0].data_res
        x = x_data
        y = y_data * var.visual.axis[1].data_scale
        if y_err_data is None:
            y_err = np.empty_like(y)
            y_err[:] = 0
        else:
            y_err = y_err_data * var.visual.axis[1].data_scale

        # resample time if needed
        x_data = x
        y_data = y
        y_err_data = y_err
        time_gap = var.visual.axis[0].mask_gap
        if time_gap is None:
            time_gap = self.time_gap
        if time_gap:
            x, y = arraytool.data_resample(
                x_data, y_data, xtype='datetime', xres=x_data_res, method='Null', axis=0)

            x, y_err = arraytool.data_resample(
                x_data, y_err_data, xtype='datetime', xres=x_data_res, method='Null', axis=0)
        data = {'x': x, 'y': y, 'y_err': y_err}
        return data

    def _retrieve_data_2d(self, var):
        x_data = var.get_visual_axis_attr(axis=0, attr_name='data')
        if type(x_data) == list:
            x_data = x_data[0]
        if x_data is None:
            depend_0 = var.get_depend(axis=0)
            x_data = depend_0['UT']  # numpy array, type=datetime
            if x_data is None:
                x_data = np.array([self._xlim[0], self._xlim[1]]).reshape(2, 1)
                var.visual.axis[0].mask_gap = False

        y_data = var.get_visual_axis_attr(axis=1, attr_name='data')
        if type(y_data) == list:
            y_data = y_data[0]
        if y_data is None:
            y_data = var.get_depend(axis=1, retrieve_data=True)
            y_data_keys = list(y_data.keys())
            y_data = y_data[y_data_keys[0]]
            if y_data is None:
                y_data = np.array([0, 1]).reshape(1, 2)
        y_data = y_data * var.visual.axis[1].data_scale

        z_data = var.get_visual_axis_attr(axis=2, attr_name='data')
        if type(z_data) == list:
            z_data = z_data[0]
        if z_data is None:
            z_data = var.value
            if z_data is None:
                z_data = np.array([[np.nan, np.nan], [np.nan, np.nan]])
        if (x_data is None) or (z_data is None):
            raise ValueError

        z_data = z_data * var.visual.axis[2].data_scale
        x_data_res = var.visual.axis[0].data_res
        time_gap = var.visual.axis[0].mask_gap
        if time_gap is None:
            time_gap = self.time_gap
        if time_gap:
            x, y, z = arraytool.data_resample_2d(
                x=x_data, y=y_data, z=z_data, xtype='datetime', xres=x_data_res, method='Null', axis=0)
        else:
            x = x_data
            y = y_data
            z = z_data

        data = {'x': x, 'y': y, 'z': z}
        return data

    @staticmethod
    def generate_label(label: str, unit: str='', style: str='double'):
        label = label
        if str(unit):
            if style == 'single':
                label = label + " " + '(' + unit + ')'
            elif style == 'double':
                label = label + "\n" + '(' + unit + ')'
            else:
                raise NotImplementedError
        return label


def check_panel_ax(func):
    def wrapper(*args, **kwargs):
        obj = args[-1]
        kwargs.setdefault('ax', None)
        if kwargs['ax'] is None:
            kwargs['ax'] = obj.axes['major']
        result = func(*args, **kwargs)
        return result
    return wrapper


class Panel1(object):
    def __init__(self, *args, figure=None, **kwargs):
        if figure is None:
            figure = plt.gcf()
        self.figure = figure
        self.axes = {}
        self.label = kwargs.pop('label', None)
        # self.objectives = kwargs.pop('objectives', {})
        ax = self.figure.add_subplot(*args, **kwargs)
        self.axes['major'] = ax

    # def add_subplot(self, *args, major=False, label=None, **kwargs):
    #     if major:
    #         label = 'major'
    #     else:
    #         if label is None:
    #             label = len(self.axes.keys())
    #
    #     ax = self.figure.add_subplot(*args, **kwargs)
    #     self.axes[label] = ax
    #     return ax

    def add_axes(self, *args, major=False, label=None, **kwargs):
        if major:
            label = 'major'
        else:
            if label is None:
                label = len(self.axes.keys())

        ax = self.figure.add_axes(*args, **kwargs)
        ax.patch.set_alpha(0)
        self.axes[label] = ax
        return ax

    @property
    def major_ax(self):
        return self.axes['major']

    @major_ax.setter
    def major_ax(self, ax):
        self.axes['major'] = ax

    @check_panel_ax
    def plot(self, *args, ax=None, **kwargs):
        # plot_type "1P"
        ipl = ax.plot(*args, **kwargs)
        return ipl

    @check_panel_ax
    def errorbar(self, *args, ax=None, **kwargs):
        # plot_type = "1E"
        ieb = ax.errorbar(*args, **kwargs)
        return ieb

    @check_panel_ax
    def pcolormesh(self, *args, ax=None, **kwargs):
        # plot_type: "2P"
        ipm = ax.pcolormesh(*args, **kwargs)
        return ipm

    @check_panel_ax
    def imshow(self, *args, ax=None, **kwargs):
        # plot_type = "2I"
        im = ax.imshow(*args, **kwargs)
        return im

    def add_label(self, x, y, label, ax=None, ha='left', va='center', **kwargs):
        if ax is None:
            ax = self.axes['major']
        if label is None:
            label = ''
        transform = kwargs.pop('transform', ax.transAxes)
        ax.text(x, y, label, transform=transform, ha=ha, va=va, **kwargs)

    def add_title(self, *args, **kwargs):
        self.axes['major'].set_title(*args, **kwargs)


def test_tspanel():
    from geospacelab.express.eiscat_dashboard import EISCATDashboard
    dt_fr = datetime.datetime.strptime('20211010' + '1700', '%Y%m%d%H%M')
    dt_to = datetime.datetime.strptime('20211010' + '2100', '%Y%m%d%H%M')

    site = 'UHF'
    antenna = 'UHF'
    modulation = 'ant'
    load_mode = 'AUTO'
    data_file_type = 'eiscat-hdf5'

    viewer = EISCATDashboard(dt_fr, dt_to, site=site, antenna=antenna, modulation=modulation,
                                 data_file_type=data_file_type, load_mode=load_mode, status_control=True,
                                 residual_control=True)

    # select beams before assign the variables
    # viewer.dataset.select_beams(field_aligned=True)
    # viewer.dataset.select_beams(az_el_pairs=[(188.6, 77.7)])
    viewer.status_mask()

    n_e = viewer.assign_variable('n_e')
    T_i = viewer.assign_variable('T_i')
    T_e = viewer.assign_variable('T_e')
    v_i = viewer.assign_variable('v_i_los')
    az = viewer.assign_variable('AZ')
    el = viewer.assign_variable('EL')

    panel1 = TSPanel(dt_fr=dt_fr, dt_to=dt_to)
    panel1.draw([n_e])
    plt.show()


if __name__ == '__main__':
    test_tspanel()


# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

import datetime
import requests
import bs4
import os
import zlib

import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.toolbox.utilities.pylogging as mylog
from geospacelab import preferences as pfr


class Downloader(object):
    """
    A class to Download SSUSI data
    :param file_type:  "l1b", "edr-aur", or "sdr"
    """
    def __init__(self, dt_fr, dt_to, sat_id=None, orbit_id=None, data_file_root_dir=None, file_type='edr_aur'):

        dt_fr = dttool.get_start_of_the_day(dt_fr)
        dt_to = dttool.get_start_of_the_day(dt_to)
        self.file_type = file_type
        if dt_fr == dt_to:
            dt_to = dt_to + datetime.timedelta(hours=23, minutes=59)
        self.dt_fr = dt_fr  # datetime from
        self.dt_to = dt_to  # datetime to
        self.sat_id = sat_id.lower()
        self.orbit_id = orbit_id

        if data_file_root_dir is None:
            self.data_file_root_dir = pfr.datahub_data_root_dir / 'JHUAPL' / 'DMSP' / 'SSUSI'
        else:
            self.data_file_root_dir = data_file_root_dir
        self.done = False

        self.url_base = "https://ssusi.jhuapl.edu/"

        self.download_files()

    def download_files(self):
        """
        Get a list of the urls from input date
        and datatype and download the files
        and also move them to the corresponding
        folders.!!!
        """
        diff_days = dttool.get_diff_days(self.dt_fr, self.dt_to)
        dt0 = dttool.get_start_of_the_day(self.dt_fr)

        for iday in range(diff_days + 1):
            thisday = dt0 + datetime.timedelta(days=iday)

            # construct day of year from date
            doy = thisday.timetuple().tm_yday
            doy_str = "{:03d}".format(doy)

            payload = {"spc": self.sat_id, "type": self.file_type,
                       "Doy": doy_str, "year": "{:d}".format(thisday.year)}

            # get a list of the files from dmsp ssusi website
            # based on the data type and date
            r = requests.get(self.url_base + "data_retriver/",
                             params=payload, verify=True)
            soup = bs4.BeautifulSoup(r.text, 'html.parser')
            div_filelist = soup.find("div", {"id": "filelist"})
            href_list = div_filelist.find_all(href=True)
            url_list = [self.url_base + href["href"] for href in href_list]

            import natsort
            url_list = natsort.natsorted(url_list, reverse=False)

            for f_url in url_list:

                # we only need data files which have .NC
                if ".NC" not in f_url:
                    continue
                # If working with sdr data use only
                # sdr-disk files
                if self.file_type == "sdr":
                    if "SDR-DISK" not in f_url:
                        continue

                if self.orbit_id is not None:
                    if len(self.orbit_id) != 5:
                        raise ValueError
                    if self.orbit_id not in f_url:
                        continue

                file_dir = self.data_file_root_dir / self.sat_id.lower() / thisday.strftime("%Y%m%d")
                file_dir.mkdir(parents=True, exist_ok=True)
                file_name = f_url.split('/')[-1]
                file_path = file_dir / file_name
                if file_path.is_file():
                    self.done = True
                    mylog.simpleinfo.info("The file {} exists.".format(file_name))
                    continue
                mylog.simpleinfo.info("Downloading {} from the online database ...".format(file_name))
                rf = requests.get(f_url, verify=True)
                file_name = rf.url.split("/")[-1]

                with open(file_path, "wb") as ssusi_data:
                    ssusi_data.write(rf.content)
                mylog.simpleinfo.info("Done. The file has been saved to {}".format(file_dir))
                self.done = True
                if self.orbit_id is not None:
                    return
                # self.file_dir = file_dir
                # self.file_name = file_name
            if self.orbit_id is None and self.done:
                fp_log = file_dir / (self.file_type.upper() + '.full.log')
                fp_log.touch()
            if not self.done:
                mylog.StreamLogger.warning(
                    "Cannot find the requested data on {} from the online database!".format(thisday.strftime("%Y-%m-%d")))


if __name__ == "__main__":
    dt_fr = datetime.datetime(2015, 12, 5)
    dt_to = datetime.datetime(2015, 12, 6)
    file_type = 'edr-aur'
    sat_id = 'f17'
    orbit_id = '46871'
    orbit_id = None

    Downloader(dt_fr, dt_to, sat_id=sat_id, orbit_id=orbit_id, file_type=file_type)


import datetime
import matplotlib.pyplot as plt

import geospacelab.visualization.mpl.dashboards as dashboards


def test():
    dt_fr = datetime.datetime(2016, 3, 14, 0)
    dt_to = datetime.datetime(2016, 3, 15, 23,)
    dashboard = dashboards.TSDashboard(dt_fr=dt_fr, dt_to=dt_to)
    ds1 = dashboard.dock(
        datasource_contents=['madrigal', 'isr', 'millstonehill', 'basic'], data_file_type='combined',
        antenna='misa', pulse_code='alternating', pulse_length=480)
    ds1.select_beams(az_el_pairs=[[-13.5, 45]])
    n_e = dashboard.assign_variable('n_e')
    T_e = dashboard.assign_variable('T_e')
    T_i = dashboard.assign_variable('T_i')
    v_i_los = dashboard.assign_variable('v_i_los')
    az = dashboard.assign_variable('AZ')
    el = dashboard.assign_variable('EL')

    ds2 = dashboard.dock(
        datasource_contents=['madrigal', 'isr', 'millstonehill', 'gridded'])
    layout = [[n_e], [T_e], [T_i], [v_i_los], [az, [el]]]
    # layout = [[Bz, By], [v_sw], [n_p], [sym_h]]
    dashboard.set_layout(panel_layouts=layout, hspace=0.1)
    dashboard.draw()
    # dashboard.save_figure(file_name='example_omni_6', append_time=False)
    pass


if __name__ == "__main__":
    test()
    plt.show()



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import datetime
import numpy as np

from geospacelab.cs._cs_base import SpaceCSBase, SphericalCoordinates, CartesianCoordinates
import geospacelab.toolbox.utilities.pylogging as mylog


class APEX(SpaceCSBase):
    def __init__(self, coords=None, ut=None, **kwargs):
        kwargs.setdefault('new_coords', ['lat', 'lon', 'r', 'mlt'])
        super().__init__(name='APEX', coords=coords, ut=ut, kind='sph', **kwargs)



import datetime
import matplotlib.pyplot as plt

import geospacelab.visualization.mpl.geomap.geodashboards as geomap

dt_fr = datetime.datetime(2015, 9, 8, 8)
dt_to = datetime.datetime(2015, 9, 8, 23, 59)
time1 = datetime.datetime(2015, 9, 8, 20, 21)
pole = 'N'
sat_id = 'f16'
band = 'LBHS'

# Create a geodashboard object
dashboard = geomap.GeoDashboard(dt_fr=dt_fr, dt_to=dt_to, figure_config={'figsize': (5, 5)})

# If the orbit_id is specified, only one file will be downloaded. This option saves the downloading time.
# dashboard.dock(datasource_contents=['jhuapl', 'dmsp', 'ssusi', 'edraur'], pole='N', sat_id='f17', orbit_id='46863')
# If not specified, the data during the whole day will be downloaded.
dashboard.dock(datasource_contents=['jhuapl', 'dmsp', 'ssusi', 'edraur'], pole=pole, sat_id=sat_id, orbit_id=None)
ds_s1 = dashboard.dock(
    datasource_contents=['madrigal', 'satellites', 'dmsp', 's1'],
    dt_fr=time1 - datetime.timedelta(minutes=45),
    dt_to=time1 + datetime.timedelta(minutes=45),
    sat_id=sat_id)

dashboard.set_layout(1, 1)

# Get the variables: LBHS emission intensiy, corresponding times and locations
lbhs = dashboard.assign_variable('GRID_AUR_' + band, dataset_index=1)
dts = dashboard.assign_variable('DATETIME', dataset_index=1).value.flatten()
mlat = dashboard.assign_variable('GRID_MLAT', dataset_index=1).value
mlon = dashboard.assign_variable('GRID_MLON', dataset_index=1).value
mlt = dashboard.assign_variable(('GRID_MLT'), dataset_index=1).value

# Search the index for the time to plot, used as an input to the following polar map
ind_t = dashboard.datasets[1].get_time_ind(ut=time1)
lbhs_ = lbhs.value[ind_t, :, :]
mlat_ = mlat[ind_t, ::]
mlon_ = mlon[ind_t, ::]
mlt_ = mlt[ind_t, ::]
# Add a polar map panel to the dashboard. Currently the style is the fixed MLT at mlt_c=0. See the keywords below:
panel1 = dashboard.add_polar_map(row_ind=0, col_ind=0, style='mlt-fixed', cs='AACGM', mlt_c=0., pole=pole, ut=time1, boundary_lat=65., mirror_south=True)

# Some settings for plotting.
pcolormesh_config = lbhs.visual.plot_config.pcolormesh
# Overlay the SSUSI image in the map.
ipm = panel1.overlay_pcolormesh(data=lbhs_, coords={'lat': mlat_, 'lon': mlon_, 'mlt': mlt_}, cs='AACGM',
                                regridding=True, **pcolormesh_config)
# Add a color bar
panel1.add_colorbar(ipm, c_label=band + " (R)", c_scale=pcolormesh_config['c_scale'], left=1.1, bottom=0.1,
                    width=0.05, height=0.7)

# Overlay the gridlines
panel1.overlay_gridlines(lat_res=5, lon_label_separator=5)

# Overlay the coastlines in the AACGM coordinate
panel1.overlay_coastlines()

# Overlay cross-track velocity along satellite trajectory
sc_dt = ds_s1['SC_DATETIME'].value.flatten()
sc_lat = ds_s1['SC_GEO_LAT'].value.flatten()
sc_lon = ds_s1['SC_GEO_LON'].value.flatten()
sc_alt = ds_s1['SC_GEO_ALT'].value.flatten()
sc_coords = {'lat': sc_lat, 'lon': sc_lon, 'height': sc_alt}

v_H = ds_s1['v_i_H'].value.flatten()
panel1.overlay_cross_track_vector(vector=v_H, unit_vector=1000, alpha=0.5, color='r', sc_coords=sc_coords, cs='GEOC', sc_ut=sc_dt)
# Overlay the satellite trajectory with ticks
panel1.overlay_sc_trajectory(sc_ut=sc_dt, sc_coords=sc_coords, cs='GEOC')

# Add the title and save the figure
polestr = 'North' if pole == 'N' else 'South'
panel1.add_title(title='DMSP/SSUSI, ' + band + ', ' + sat_id.upper() + ', ' + polestr + ', ' + time1.strftime('%Y-%m-%d %H%M UT'))
plt.savefig('DMSP_SSUSI_' + time1.strftime('%Y%m%d-%H%M') + '_' + band + '_' + sat_id.upper() + '_' + pole, dpi=300)

# show the figure
plt.show()


import datetime
import matplotlib.pyplot as plt
import numpy as np

import geospacelab.visualization.mpl.geomap.geodashboards as geomap


def test_ampere():
    dt_fr = datetime.datetime(2016, 3, 15, 0)
    dt_to = datetime.datetime(2016, 3, 15, 23, 59)
    time1 = datetime.datetime(2016, 3, 15, 1, 10)
    pole = 'N'
    load_mode = 'assigned'
    # specify the file full path
    data_file_paths = ['/home/lei/afys-data/SuperDARN/PotentialMap/2016/test.dat']
    # data_file_paths = ['/Users/lcai/Geospacelab/Data/SuperDARN/POTMAP/2016/SuperDARM_POTMAP_20160314_10min_test.txt']

    viewer = geomap.GeoDashboard(dt_fr=dt_fr, dt_to=dt_to, figure_config={'figsize': (8, 8)})
    viewer.dock(datasource_contents=['superdarn', 'potmap'], load_mode=load_mode, data_file_paths=data_file_paths)
    viewer.set_layout(1, 1)
    dataset_superdarn = viewer.datasets[1]

    phi = viewer.assign_variable('GRID_phi', dataset_index=1)
    dts = viewer.assign_variable('DATETIME', dataset_index=1).value.flatten()
    mlat = viewer.assign_variable('GRID_MLAT', dataset_index=1)
    mlon = viewer.assign_variable('GRID_MLON', dataset_index=1)
    mlt = viewer.assign_variable(('GRID_MLT'), dataset_index=1)

    ind_t = dataset_superdarn.get_time_ind(ut=time1)
    # initialize the polar map
    panel1 = viewer.add_polar_map(row_ind=0, col_ind=0, style='mlt-fixed', cs='AACGM', mlt_c=0., pole=pole, ut=time1, boundary_lat=50, mirror_south=True)

    panel1.overlay_coastlines()

    phi_ = phi.value[ind_t]
    mlat_ = mlat.value[ind_t]
    mlt_ = mlt.value[ind_t]
    mlon_ = mlon.value[ind_t]

    # grid_mlat, grid_mlt, grid_phi = dataset_superdarn.grid_phi(mlat_, mlt_, phi_, interp_method='cubic')
    grid_mlat, grid_mlt, grid_phi = dataset_superdarn.postprocess_roll(mlat_, mlt_, phi_)

    # re-grid the original data with higher spatial resolution, default mlt_res = 0.05, mlat_res = 0.5. used for plotting.
    # grid_mlat, grid_mlt, grid_fac = dataset_ampere.grid_fac(phi_, mlt_res=0.05, mlat_res=0.05, interp_method='linear')

    levels = np.array([-21e3, -18e3, -15e3, -12e3, -9e3, -6e3, 3e3, 6e3,  9e3, 12e3, 15e3, 18e3, 21e3])
    # ipc = panel1.add_pcolor(fac_, coords={'lat': mlat[ind_t, ::], 'lon': None, 'mlt': mlt[ind_t, ::], 'height': 250.}, cs='AACGM', **pcolormesh_config)
    ict = panel1.overlay_contour(grid_phi, coords={'lat': grid_mlat, 'lon': None, 'mlt': grid_mlt}, cs='AACGM', colors='b', levels=levels)
    # panel1.major_ax.clabel(ict, inline=True, fontsize=10)
    panel1.overlay_gridlines(lat_res=5, lon_label_separator=5)

    polestr = 'North' if pole == 'N' else 'South'
    # panel1.add_title('DMSP/SSUSI, ' + band + ', ' + sat_id.upper() + ', ' + polestr + ', ' + time1.strftime('%Y-%m-%d %H%M UT'), pad=20)
    plt.savefig('superdarn_example', dpi=300)
    plt.show()


if __name__ == "__main__":
    test_ampere()



import datetime
import matplotlib.pyplot as plt

import geospacelab.express.eiscat_dashboard as eiscat


def test_UHF_CP2():
    dt_fr = datetime.datetime.strptime('20161023' + '1600', '%Y%m%d%H%M')
    dt_to = datetime.datetime.strptime('20161023' + '2100', '%Y%m%d%H%M')

    site = 'UHF'
    antenna = 'UHF'
    modulation = '44.4'
    load_mode = 'AUTO'
    data_file_type = 'eiscat-hdf5'

    dashboard = eiscat.EISCATDashboard(dt_fr, dt_to, site=site, antenna=antenna, modulation=modulation,
                                 data_file_type=data_file_type, load_mode=load_mode, status_control=True,
                                 residual_control=True)

    # select beams before assign the variables
    # dashboard.dataset.select_beams(field_aligned=True)
    dashboard.dataset.select_beams(az_el_pairs=[(187.0, 77.7)])
    dashboard.check_beams()

    dashboard.status_mask()

    n_e = dashboard.assign_variable('n_e')
    T_i = dashboard.assign_variable('T_i')
    T_e = dashboard.assign_variable('T_e')
    v_i = dashboard.assign_variable('v_i_los')
    v_i_2 = v_i.clone()
    v_i_2.value = v_i_2.value
    v_i.visual.axis[2].lim=[-100, 100]
    az = dashboard.assign_variable('AZ')
    el = dashboard.assign_variable('EL')
    ptx = dashboard.assign_variable('P_Tx')
    tsys = dashboard.assign_variable('T_SYS_1')
    # T_r = dashboard.dataset.add_variable('T_r', ndim=2)
    # T_r.value = T_e.value / T_i.value
    # T_r.visual.plot_config.style = '2P'
    # T_r.depends = T_i.depends
    # T_r.visual.axis[2].label = 'Te/Ti'
    # T_r.visual.axis[2].unit = ''

    layout = [[n_e], [T_e], [T_i], [v_i], [az, [el], [ptx], [tsys]]]
    dashboard.set_layout(panel_layouts=layout, )
    dashboard.draw()
    dashboard.add_title()
    dashboard.add_panel_labels()
    return dashboard


def test_esr_32m():
    dt_fr = datetime.datetime.strptime('20001201' + '0400', '%Y%m%d%H%M')
    dt_to = datetime.datetime.strptime('20001201' + '1200', '%Y%m%d%H%M')

    site = 'ESR'
    antenna = '42m'
    modulation = '120'
    load_mode = 'AUTO'
    data_file_type = 'madrigal-hdf5'

    dashboard = eiscat.EISCATDashboard(dt_fr, dt_to, site=site, antenna=antenna, modulation=modulation,
                                 data_file_type=data_file_type, load_mode=load_mode)

    # select beams before assign the variables
    # dashboard.dataset.select_beams(field_aligned=False)

    n_e = dashboard.assign_variable('n_e')
    T_i = dashboard.assign_variable('T_i')
    T_e = dashboard.assign_variable('T_e')
    v_i = dashboard.assign_variable('v_i_los')
    az = dashboard.assign_variable('AZ')
    el = dashboard.assign_variable('EL')

    layout = [[n_e], [T_e], [T_i], [v_i], [az, el]]
    dashboard.set_layout(panel_layouts=layout, row_height_scales=[5, 5, 5, 5, 3])
    dashboard.draw()
    dashboard.add_title()
    dashboard.add_panel_labels()
    return dashboard


def test_uhf_cp3():
    dt_fr = datetime.datetime.strptime('20210322' + '2000', '%Y%m%d%H%M')
    dt_to = datetime.datetime.strptime('20210322' + '2300', '%Y%m%d%H%M')

    site = 'UHF'
    antenna = 'UHF'
    modulation = 'ant'
    load_mode = 'AUTO'
    data_file_type = 'madrigal-hdf5'

    dashboard = eiscat.EISCATDashboard(dt_fr, dt_to, site=site, antenna=antenna, modulation=modulation,
                                 data_file_type=data_file_type, load_mode=load_mode)
    dashboard.check_beams()
    # select beams before assign the variables
    # dashboard.dataset.select_beams(field_aligned=False)

    n_e = dashboard.assign_variable('n_e')

    n_e.visual.axis[0].data_res = 180 # time resolution in seconds
    n_e.visual.axis[1].data = '@d.GEO_LAT.value'
    n_e.visual.axis[1].label = 'GLAT'
    n_e.visual.axis[1].unit = 'deg'
    n_e.visual.axis[1].lim = [65, 75]
    T_i = dashboard.assign_variable('T_i')
    T_i.visual.axis[0].data_res = 180
    T_e = dashboard.assign_variable('T_e')
    T_e.visual.axis[0].data_res = 180

    v_i = dashboard.assign_variable('v_i_los')
    v_i.visual.axis[0].data_res = 180
    v_i.visual.axis[1].data = '@d.GEO_LAT.value'
    v_i.visual.axis[1].label = 'GLAT'
    v_i.visual.axis[1].unit = 'deg'
    v_i.visual.axis[1].lim = [65, 75]

    az = dashboard.assign_variable('AZ')
    az.visual.axis[0].data_res = 180
    el = dashboard.assign_variable('EL')
    el.visual.axis[0].data_res = 180

    layout = [[n_e], [T_e], [T_i], [v_i], [az, el]]
    dashboard.set_layout(panel_layouts=layout, row_height_scales=[5, 5, 5, 5, 3])
    dashboard.draw()
    dashboard.add_title()
    dashboard.add_panel_labels()
    return dashboard


def test_vhf_lowel():
    dt_fr = datetime.datetime.strptime('20060903' + '2030', '%Y%m%d%H%M')
    dt_to = datetime.datetime.strptime('20060903' + '2359', '%Y%m%d%H%M')

    site = 'VHF'
    antenna = 'VHF'
    modulation = '60'
    load_mode = 'AUTO'
    data_file_type = 'madrigal-hdf5'

    dashboard = eiscat.EISCATDashboard(dt_fr, dt_to, figure_config={'figsize': (30, 8)}, site=site, antenna=antenna, modulation=modulation,
                                 data_file_type=data_file_type, load_mode=load_mode)
    # select beams before assign the variables
    # dashboard.dataset.select_beams(field_aligned=False)

    n_e = dashboard.assign_variable('n_e')
    n_e_clone = n_e.clone()         # Clone a variable. It has no meaning here, but it may be useful to create panels with the same variable but different visual settings.
    n_e.visual.axis[0].data_res = 180 # time resolution in seconds
    n_e.visual.axis[1].data = '@d.AACGM_LAT.value'
    n_e.visual.axis[1].label = 'MALT'
    n_e.visual.axis[1].unit = 'deg'
    n_e.visual.axis[1].lim = [68, 74.5]
    T_i = dashboard.assign_variable('T_i')
    T_i.visual.axis[0].data_res = 180 # time resolution in seconds
    T_i.visual.axis[1].data = '@d.GEO_LAT.value'
    T_i.visual.axis[1].label = 'GLAT'
    T_i.visual.axis[1].unit = 'deg'
    T_i.visual.axis[1].lim = [71, 75]
    T_e = dashboard.assign_variable('T_e')
    T_e.visual.axis[0].data_res = 180 # time resolution in seconds
    T_e.visual.axis[1].data = '@d.GEO_LAT.value'
    T_e.visual.axis[1].label = 'GLAT'
    T_e.visual.axis[1].unit = 'deg'
    T_e.visual.axis[1].lim = [71, 75]
    v_i = dashboard.assign_variable('v_i_los')
    v_i.visual.axis[0].data_res = 180 # time resolution in seconds
    v_i.visual.axis[1].data = '@d.GEO_LAT.value'
    v_i.visual.axis[1].label = 'GLAT'
    v_i.visual.axis[1].unit = 'deg'
    v_i.visual.axis[1].lim = [None, None]
    az = dashboard.assign_variable('AZ')
    el = dashboard.assign_variable('EL')

    layout = [[n_e], [T_e], [T_i], [v_i], [az, el]]
    dashboard.set_layout(panel_layouts=layout, row_height_scales=[5, 5, 5, 5, 3], right=0.4)
    dashboard.draw()
    dashboard.add_title()
    dashboard.add_panel_labels()

    dt_fr = datetime.datetime.strptime('20001201' + '0400', '%Y%m%d%H%M')
    dt_to = datetime.datetime.strptime('20001201' + '1200', '%Y%m%d%H%M')

    site = 'ESR'
    antenna = '42m'
    modulation = '120'
    load_mode = 'AUTO'
    data_file_type = 'madrigal-hdf5'

    dashboard = dashboard.figure.add_dashboard(dt_fr=dt_fr, dt_to=dt_to, site=site, antenna=antenna, modulation=modulation,
                                       data_file_type=data_file_type, load_mode=load_mode)

    # select beams before assign the variables
    # dashboard.dataset.select_beams(field_aligned=False)

    n_e = dashboard.assign_variable('n_e')
    T_i = dashboard.assign_variable('T_i')
    T_e = dashboard.assign_variable('T_e')
    v_i = dashboard.assign_variable('v_i_los')
    az = dashboard.assign_variable('AZ')
    el = dashboard.assign_variable('EL')

    layout = [[n_e], [T_e], [T_i], [v_i], [az, el]]
    dashboard.set_layout(panel_layouts=layout, row_height_scales=[5, 5, 5, 5, 3], left=0.6)
    dashboard.draw()
    dashboard.add_title()
    dashboard.add_panel_labels()

    return dashboard


if __name__ == "__main__":
    # db = test_vhf_lowel()

    # test_uhf_cp3()
    test_UHF_CP2()
    # test_esr_32m()

    plt.savefig('eiscat_example.png')
    plt.show()


# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"

from geospacelab.visualization.mpl.__base__ import FigureBase
from geospacelab.visualization.mpl.dashboards import Dashboard
from geospacelab.visualization.mpl.panels import Panel

def create_figure(*args, watermark=None, watermark_style=None, **kwargs):
    return FigureBase(*args, watermark=None, watermark_style=None, **kwargs)




# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"

import numpy as np
import copy
from geospacelab.datahub import VariableModel as Var
import geospacelab.visualization.mpl.colormaps as cm

database = 'Madrigal'

timestamps = {
    'DATETIME': 'SC_DATETIME',
}


coords = {
    'GEO': ['SC_GEO_LAT', 'SC_GEO_LON', 'SC_GEO_ALT', 'SC_GEO_ST'],
    'AACGM': ['SC_AACGM_LAT', 'SC_AACGM_LON', 'SC_AACGM_R', 'SC_AACGM_MLT'],
    'APEX': ['SC_APEX_LAT', 'SC_APEX_LON', 'SC_APEX_ALT', 'SC_APEX_MLT']
}

depend_0 = {
    'UT': 'SC_DATETIME',
    'GEO_LAT': 'SC_GEO_LAT', 'GEO_LON': 'SC_GEO_LON',
    'AACGM_LAT': 'SC_AACGM_LAT', 'AACGM_LON': 'SC_AACGM_LON', 'AACGM_MLT': 'SC_AACGM_MLT'
}

depend_1 = {'ENERGY_CHANNEL': 'ENERGY_CHANNEL_GRID'}

default_colormap = cm.cmap_jet_modified()

default_axis_dict_1d = {
    1: {
        'data':     '@v.value',
        'label':    '@v.group',
        'unit':     '@v.unit',
        'label_pos': [-0.1, 0.5],
    },
    2: {
        'label':    '@v.label'
    },
}

default_axis_dict_2d = {
    1:     {
        'data':     '@d.ENERGY_CHANNEL_GRID.value',
        'lim':      [2.5e1, 4e4],
        'scale':    'log',
        'label':    'Energy',
        'unit':     'eV',
        'label_pos': [-0.1, 0.5],
    },
    2:  {
        'data':     '@v.value',
        'label':    '@v.label',
        'unit':     '@v.unit_label',
        'scale':    'log'
    }
}

default_plot_config = {
    'line':         {
        'linestyle':        '',
        'linewidth':        1.5,
        'marker':           '.',
        'markersize':       3,
    },
    'pcolormesh':   {
        'cmap':            default_colormap,
    }
}

configured_variables = {}
visual = 'on'

####################################################################################################################
var_name = 'JE_e'
var = Var(ndim=1, variable_type='scalar', visual=visual)
# set variable attrs
var_config = {
    'name': var_name,
    'fullname': 'Integrated energy flux of electrons',
    'label': r'JE$_e$',
    'group': 'JE',
    'unit': r'eV/cm2/s/ster',
    'unit_label': r'eV$/$cm$^{2}/$s$/$ster',
    'error': None,
    'depends': {0: depend_0},
}
var.config(**var_config)
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '1noE'
# set axis attrs
var.visual.axis[1].config(**default_axis_dict_1d[1])
var.visual.axis[1].lim = [2e7, 3e13]
var.visual.axis[1].scale = 'log'

var.visual.axis[2].config(**default_axis_dict_1d[2])

configured_variables[var_name] = var

####################################################################################################################
var_name = 'JE_i'
var = Var(ndim=1, variable_type='scalar', visual=visual)
# set variable attrs
var_config = {
    'name': var_name,
    'fullname': 'Integrated energy flux of ions',
    'label': r'JE$_i$',
    'group': 'JE',
    'unit': r'eV/cm2/s/ster',
    'unit_label': r'eV$/$cm$^{2}/$s$/$ster',
    'error': None,
    'depends': {0: depend_0},
}
var.config(**var_config)
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '1noE'
# set axis attrs
var.visual.axis[1].config(**default_axis_dict_1d[1])
var.visual.axis[1].lim = [2e7, 3e13]
var.visual.axis[1].scale = 'log'

var.visual.axis[2].config(**default_axis_dict_1d[2])

configured_variables[var_name] = var

####################################################################################################################
var_name = 'JN_e'
var = Var(ndim=1, variable_type='scalar', visual=visual)
# set variable attrs
var_config = {
    'name': var_name,
    'fullname': 'Integrated number flux of electrons',
    'label': r'JN$_e$',
    'group': 'JN',
    'unit': r'#/cm2/s/ster',
    'unit_label': r'$\#/$cm$^{2}/$s$/$ster',
    'error': None,
    'depends': {0: depend_0},
}
var.config(**var_config)
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '1noE'
# set axis attrs
var.visual.axis[1].config(**default_axis_dict_1d[1])
var.visual.axis[1].lim = [1e4, 3e9]
var.visual.axis[1].scale = 'log'

var.visual.axis[2].config(**default_axis_dict_1d[2])

configured_variables[var_name] = var

####################################################################################################################
var_name = 'JN_i'
var = Var(ndim=1, variable_type='scalar', visual=visual)
# set variable attrs
var_config = {
    'name': var_name,
    'fullname': 'Integrated number flux of ions',
    'label': r'JN$_e$',
    'group': 'JN',
    'unit': r'#/cm2/s/ster',
    'unit_label': r'$\#/$cm$^{2}/$s$/$ster',
    'error': None,
    'depends': {0: depend_0},
}
var.config(**var_config)
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '1noE'
# set axis attrs
var.visual.axis[1].config(**default_axis_dict_1d[1])
var.visual.axis[1].lim = [1e4, 3e9]
var.visual.axis[1].scale = 'log'

var.visual.axis[2].config(**default_axis_dict_1d[2])

configured_variables[var_name] = var

####################################################################################################################
var_name = 'E_e_MEAN'
var = Var(ndim=1, variable_type='scalar', visual=visual)
# set variable attrs
var_config = {
    'name': var_name,
    'fullname': 'Mean energy of electrons',
    'label': r'$\bar{E}_e$',
    'group': r'$\bar{E}$',
    'unit': r'eV',
    'unit_label': None,
    'error': None,
    'depends': {0: depend_0},
}
var.config(**var_config)
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '1noE'
# set axis attrs
var.visual.axis[1].config(**default_axis_dict_1d[1])
var.visual.axis[1].lim = [2.5e1, 4e4]
var.visual.axis[1].scale = 'log'

var.visual.axis[2].config(**default_axis_dict_1d[2])

configured_variables[var_name] = var

####################################################################################################################
var_name = 'E_i_MEAN'
var = Var(ndim=1, variable_type='scalar', visual=visual)
# set variable attrs
var_config = {
    'name': var_name,
    'fullname': 'Mean energy of ions',
    'label': r'$\bar{E}_i$',
    'group': r'$\bar{E}$',
    'unit': r'eV',
    'unit_label': None,
    'error': None,
    'depends': {0: depend_0},
}
var.config(**var_config)
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '1noE'
# set axis attrs
var.visual.axis[1].config(**default_axis_dict_1d[1])
var.visual.axis[1].lim = [2.5e1, 4e4]
var.visual.axis[1].scale = 'log'

var.visual.axis[2].config(**default_axis_dict_1d[2])

configured_variables[var_name] = var

####################################################################################################################
var_name = 'jE_e'
var = Var(ndim=2, variable_type='scalar', visual=visual)
# set variable attrs
var_config = {
    'name': var_name,
    'fullname': 'Integrated number flux of electrons',
    'label': r'jE$_e$',
    'group': 'jE',
    'unit': r'eV/cm2/s/ster/Delta eV',
    'unit_label': r'eV$/$cm$^{2}/$s$/$ster$/\Delta$eV',
    'error': None,
    'depends': {0: depend_0, 1: depend_1},
}
var.config(**var_config)
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '2P'
# set axis attrs
var.visual.axis[1].config(**default_axis_dict_2d[1])

var.visual.axis[2].config(**default_axis_dict_2d[2])
var.visual.axis[2].lim = [1e5, 1e10]

configured_variables[var_name] = var

####################################################################################################################
var_name = 'jE_i'
var = Var(ndim=2, variable_type='scalar', visual=visual)
# set variable attrs
var_config = {
    'name': var_name,
    'fullname': 'Integrated number flux of ions',
    'label': r'jE$_i$',
    'group': 'jE',
    'unit': r'eV/cm2/s/ster/Delta eV',
    'unit_label': r'eV$/$cm$^{2}/$s$/$ster$/\Delta$eV',
    'error': None,
    'depends': {0: depend_0, 1: depend_1},
}
var.config(**var_config)
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '2P'
# set axis attrs
var.visual.axis[1].config(**default_axis_dict_2d[1])

var.visual.axis[2].config(**default_axis_dict_2d[2])
var.visual.axis[2].lim = [1e3, 1e8]

configured_variables[var_name] = var

####################################################################################################################
var_name = 'jN_e'
var = Var(ndim=2, variable_type='scalar', visual=visual)
# set variable attrs
var_config = {
    'name': var_name,
    'fullname': 'Integrated number flux of electrons',
    'label': r'jN$_e$',
    'group': 'jN',
    'unit': r'#/cm2/s/ster/Delta eV',
    'unit_label': r'$\#/$cm$^{2}/$s$/$ster$/\Delta$eV',
    'error': None,
    'depends': {0: depend_0, 1: depend_1},
}
var.config(**var_config)
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '2P'
# set axis attrs
var.visual.axis[1].config(**default_axis_dict_2d[1])

var.visual.axis[2].config(**default_axis_dict_2d[2])
var.visual.axis[2].lim = [1e4, 1e9]

configured_variables[var_name] = var

####################################################################################################################
var_name = 'jN_i'
var = Var(ndim=2, variable_type='scalar', visual=visual)
# set variable attrs
var_config = {
    'name': var_name,
    'fullname': 'Integrated number flux of ions',
    'label': r'jN$_i$',
    'group': 'jN',
    'unit': r'#/cm2/s/ster/Delta eV',
    'unit_label': r'$\#/$cm$^{2}/$s$/$ster$/\Delta$eV',
    'error': None,
    'depends': {0: depend_0, 1: depend_1},
}
var.config(**var_config)
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '2P'
# set axis attrs
var.visual.axis[1].config(**default_axis_dict_2d[1])

var.visual.axis[2].config(**default_axis_dict_2d[2])
var.visual.axis[2].lim = [1e3, 1e8]

configured_variables[var_name] = var



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import datetime

from geospacelab.visualization.mpl.dashboards import TSDashboard
import geospacelab.toolbox.utilities.pylogging as mylog


class DMSPDashboard(TSDashboard):
    def __init__(
            self,
            dt_fr: datetime.datetime,
            dt_to: datetime.datetime,
            sat_id: str = None,
            load_mode: str = 'AUTO',
            figure: str = 'new',
            figure_config: dict = {'figsize': (10, 10)},
            timeline_extra_labels: list = None,
            **kwargs
    ):
        if timeline_extra_labels is None:
            timeline_extra_labels = ['GEO_LAT', 'GEO_LON', 'AACGM_LAT', 'AACGM_MLT']
        super().__init__(dt_fr=dt_fr, dt_to=dt_to,
                         figure=figure, figure_config=figure_config,
                         timeline_extra_labels=timeline_extra_labels)
        self.dataset_s1 = self.dock(datasource_contents=['madrigal', 'dmsp', 's1'], sat_id=sat_id, load_mode=load_mode, **kwargs)
        self.dataset_e = self.dock(datasource_contents=['madrigal', 'dmsp', 'e'], sat_id=sat_id, load_mode=load_mode, **kwargs)
        self.dataset_s4 = self.dock(datasource_contents=['madrigal', 'dmsp', 's4'], sat_id=sat_id, load_mode=load_mode, **kwargs)
        # ds_1.list_all_variables()
        self.title = kwargs.pop('title', ', '.join([self.dataset_s1.database, self.dataset_s1.facility, self.dataset_s1.sat_id]))

    def save_figure(self, **kwargs):
        file_name = kwargs.pop('file_name', self.title.replace(', ', '_'))
        super().save_figure(file_name=file_name, **kwargs)

    def add_title(self, **kwargs):
        title = kwargs.pop('title', self.title)
        super().add_title(x=0.5, y=1.06, title=title)

    def quicklook(self):

        n_e = self.assign_variable('n_e', dataset=self.dataset_s1)
        v_i_H = self.assign_variable('v_i_H', dataset=self.dataset_s1)
        v_i_V = self.assign_variable('v_i_V', dataset=self.dataset_s1)
        d_B_D = self.assign_variable('d_B_D', dataset=self.dataset_s1)
        d_B_P = self.assign_variable('d_B_P', dataset=self.dataset_s1)
        d_B_F = self.assign_variable('d_B_F', dataset=self.dataset_s1)

        JE_e = self.assign_variable('JE_e', dataset=self.dataset_e)
        JE_i = self.assign_variable('JE_i', dataset=self.dataset_e)
        jE_e = self.assign_variable('jE_e', dataset=self.dataset_e)
        jE_i = self.assign_variable('jE_i', dataset=self.dataset_e)
        E_e_MEAN = self.assign_variable('E_e_MEAN', dataset=self.dataset_e)
        E_i_MEAN = self.assign_variable('E_i_MEAN', dataset=self.dataset_e)

        T_i = self.assign_variable('T_i', dataset=self.dataset_s4)
        T_e = self.assign_variable('T_e', dataset=self.dataset_s4)
        c_O_p = self.assign_variable('COMP_O_p', dataset=self.dataset_s4)

        self.list_assigned_variables()
        self.list_datasets()

        layout = [
            [v_i_H, v_i_V],
            [d_B_P, d_B_D, d_B_F],
            [E_e_MEAN, E_i_MEAN],
            [JE_e, JE_i],
            [jE_e],
            [jE_i],
            [n_e, [c_O_p]],
            [T_e, T_i],
        ]
        self.set_layout(panel_layouts=layout, hspace=0.1)
        # plt.style.use('dark_background')
        # dt_fr_1 = datetime.datetime.strptime('20201209' + '1300', '%Y%m%d%H%M')
        # dt_to_1 = datetime.datetime.strptime('20201210' + '1200', '%Y%m%d%H%M')

        self.draw()
        self.add_title()
        self.add_panel_labels()


def example():

    dt_fr = datetime.datetime.strptime('20150908' + '2034', '%Y%m%d%H%M')
    dt_to = datetime.datetime.strptime('20150908' + '2036', '%Y%m%d%H%M')

    dashboard = DMSPDashboard(
        dt_fr, dt_to, sat_id='F18',
    )
    dashboard.quicklook()
    dashboard.show()


if __name__ == '__main__':
    example()




# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

from geospacelab import preferences

from geospacelab.datahub import DatabaseModel


class JHUAPLDatabase(DatabaseModel):
    def __new__(cls, str_in, **kwargs):
        obj = super().__new__(cls, str_in, **kwargs)
        return obj


jhuapl_database = JHUAPLDatabase('JHUAPL')
jhuapl_database.url = 'https://www.jhuapl.edu/'
jhuapl_database.category = 'online database'
jhuapl_database.Notes = ''



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

from geospacelab import preferences

from geospacelab.datahub import DatabaseModel


class ESAEODatabase(DatabaseModel):
    def __new__(cls, str_in, **kwargs):
        obj = super().__new__(cls, str_in, **kwargs)
        return obj


esaeo_database = ESAEODatabase('ESA/EarthOnline')
esaeo_database.url = 'https://earth.esa.int/'
esaeo_database.category = 'online database'
esaeo_database.Notes = ''


# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"

import numpy as np
import datetime

import geospacelab.datahub as datahub
from geospacelab.datahub import DatabaseModel, FacilityModel, InstrumentModel, ProductModel
from geospacelab.datahub.sources.madrigal import madrigal_database
from geospacelab.datahub.sources.madrigal.satellites.dmsp import dmsp_facility
from geospacelab import preferences as prf
import geospacelab.toolbox.utilities.pybasic as basic
import geospacelab.toolbox.utilities.pylogging as mylog
import geospacelab.toolbox.utilities.pydatetime as dttool
from geospacelab.datahub.sources.madrigal.satellites.dmsp.e.loader import Loader as default_Loader
from geospacelab.datahub.sources.madrigal.satellites.dmsp.downloader import Downloader as default_Downloader
import geospacelab.datahub.sources.madrigal.satellites.dmsp.e.variable_config as var_config


default_dataset_attrs = {
    'database': madrigal_database,
    'facility': dmsp_facility,
    'instrument': 'SSJ',
    'product': 'e',
    'data_file_ext': 'hdf5',
    'data_root_dir': prf.datahub_data_root_dir / 'Madrigal' / 'DMSP',
    'allow_load': True,
    'allow_download': True,
    'force_download': False,
    'data_search_recursive': False,
    'label_fields': ['database', 'facility', 'instrument', 'product'],
    'load_mode': 'AUTO',
    'time_clip': True,
    'add_AACGM': True,
    'calib_orbit': True,
    'replace_orbit': True,
}

default_variable_names = [
    'SC_DATETIME',
    'SC_GEO_LAT',
    'SC_GEO_LON',
    'SC_GEO_ALT',
    'SC_MAG_LAT',
    'SC_MAG_LON',
    'SC_MAG_MLT',
    'JN_e',
    'JN_i',
    'JE_e',
    'JE_i',
    'E_e_MEAN',
    'E_i_MEAN',
    'jE_e',
    'jE_i',
    'jN_e',
    'jN_i',
    'CH_CTRL_E',
    'ENERGY_CHANNEL_GRID',
    ]

# default_data_search_recursive = True

default_attrs_required = ['sat_id', ]


class Dataset(datahub.DatasetSourced):
    def __init__(self, **kwargs):
        kwargs = basic.dict_set_default(kwargs, **default_dataset_attrs)

        super().__init__(**kwargs)

        self.database = kwargs.pop('database', 'Madrigal')
        self.facility = kwargs.pop('facility', 'DMSP')
        self.instrument = kwargs.pop('instrument', 'SSJ')
        self.product = kwargs.pop('product', 'e')
        self.allow_download = kwargs.pop('allow_download', False)
        self.force_download = kwargs.pop('force_download', False)
        self.add_AACGM = kwargs.pop('add_AACGM', False)
        self.calib_orbit = kwargs.pop('calib_orbit', False)
        self.replace_orbit = kwargs.pop('replace_orbit', False)

        self.sat_id = kwargs.pop('sat_id', None)

        self.metadata = None

        allow_load = kwargs.pop('allow_load', False)

        # self.config(**kwargs)

        if self.loader is None:
            self.loader = default_Loader

        if self.downloader is None:
            self.downloader = default_Downloader

        self._validate_attrs()

        if allow_load:
            self.load_data()

    def _validate_attrs(self):
        for attr_name in default_attrs_required:
            attr = getattr(self, attr_name)
            if not list(attr):
                mylog.StreamLogger.warning("The parameter {} is required before loading data!".format(attr_name))

    def label(self, **kwargs):
        label = super().label()
        return label

    def load_data(self, **kwargs):
        self.check_data_files(**kwargs)

        self._set_default_variables(
            default_variable_names,
            configured_variables=var_config.configured_variables
        )
        for file_path in self.data_file_paths:
            load_obj = self.loader(file_path, file_ext='hdf5')

            for var_name in self._variables.keys():
                value = load_obj.variables[var_name]
                self._variables[var_name].join(value)

            # self.select_beams(field_aligned=True)
        if self.time_clip:
            self.time_filter_by_range(var_datetime_name='SC_DATETIME')

        if self.calib_orbit:
            self.fix_geo_lon()

        if self.add_AACGM:
            self.convert_to_AACGM()

    def convert_to_AACGM(self):
        import geospacelab.cs as gsl_cs

        coords_in = {
            'lat': self['SC_GEO_LAT'].value.flatten(),
            'lon': self['SC_GEO_LON'].value.flatten(),
            'height': self['SC_GEO_ALT'].value.flatten()
        }
        dts = self['SC_DATETIME'].value.flatten()
        cs_sph = gsl_cs.GEOCSpherical(coords=coords_in, ut=dts)
        cs_aacgm = cs_sph.to_AACGM(append_mlt=True)
        self.add_variable('SC_AACGM_LAT')
        self.add_variable('SC_AACGM_LON')
        self.add_variable('SC_AACGM_MLT')
        self['SC_AACGM_LAT'].value = cs_aacgm['lat'].reshape(self['SC_DATETIME'].value.shape)
        self['SC_AACGM_LON'].value = cs_aacgm['lon'].reshape(self['SC_DATETIME'].value.shape)
        self['SC_AACGM_MLT'].value = cs_aacgm['mlt'].reshape(self['SC_DATETIME'].value.shape)

    def fix_geo_lon(self):
        from geospacelab.observatory.sc_orbit import OrbitPosition_SSCWS
        from scipy.interpolate import interp1d
        # check outliers
        orbit_obj = OrbitPosition_SSCWS(
            dt_fr=self.dt_fr - datetime.timedelta(minutes=30),
            dt_to=self.dt_to + datetime.timedelta(minutes=30),
            sat_id='dmsp' + self.sat_id.lower()
        )

        glat_1 = self['SC_GEO_LAT'].value.flatten()
        glon_1 = self['SC_GEO_LON'].value.flatten()
        if glat_1.size < 2:
            return

        dts_1 = self['SC_DATETIME'].value.flatten()
        dt0 = dttool.get_start_of_the_day(self.dt_fr)
        sectime_1 = [(dt - dt0).total_seconds() for dt in dts_1]

        glat_2 = orbit_obj['SC_GEO_LAT'].value.flatten()
        glon_2 = orbit_obj['SC_GEO_LON'].value.flatten()
        dts_2 = orbit_obj['SC_DATETIME'].value.flatten()
        sectime_2 = [(dt - dt0).total_seconds() for dt in dts_2]

        factor = np.pi / 180.
        sin_glon_1 = np.sin(glon_1 * factor)
        sin_glon_2 = np.sin(glon_2 * factor)
        cos_glon_2 = np.cos(glon_2 * factor)
        itpf_sin = interp1d(sectime_2, sin_glon_2, kind='cubic', bounds_error=False, fill_value='extrapolate')
        itpf_cos = interp1d(sectime_2, cos_glon_2, kind='cubic', bounds_error=False, fill_value='extrapolate')
        sin_glon_2_i = itpf_sin(sectime_1)
        cos_glon_2_i = itpf_cos(sectime_1)
        rad = np.sign(sin_glon_2_i) * (np.pi / 2 - np.arcsin(cos_glon_2_i))
        glon_new = rad / factor
        # rad = np.where((rad >= 0), rad, rad + 2 * numpy.pi)

        ind_outliers = np.where(np.abs(sin_glon_1 - sin_glon_2_i) > 0.03)[0]

        if self.replace_orbit:
            glon_1 = glon_new
        else:
            glon_1[ind_outliers] = glon_new[ind_outliers]
        self['SC_GEO_LON'].value = glon_1.reshape((glon_1.size, 1))

    def search_data_files(self, **kwargs):

        dt_fr = self.dt_fr
        dt_to = self.dt_to

        diff_days = dttool.get_diff_days(dt_fr, dt_to)

        dt0 = dttool.get_start_of_the_day(dt_fr)

        for i in range(diff_days + 1):
            this_day = dt0 + datetime.timedelta(days=i)

            initial_file_dir = kwargs.pop(
                'initial_file_dir', self.data_root_dir / this_day.strftime('%Y%m') / this_day.strftime('%Y%m%d')
            )

            file_patterns = [
                'dms',
                this_day.strftime('%Y%m%d'),
                self.sat_id[1:] + self.product,
            ]
            # remove empty str
            file_patterns = [pattern for pattern in file_patterns if str(pattern)]
            search_pattern = '*' + '*'.join(file_patterns) + '*'

            done = super().search_data_files(
                initial_file_dir=initial_file_dir,
                search_pattern=search_pattern,
                allow_multiple_files=False,
            )
            # Validate file paths

            if (not done and self.allow_download) or self.force_download:
                done = self.download_data()
                if done:
                    done = super().search_data_files(
                        initial_file_dir=initial_file_dir,
                        search_pattern=search_pattern,
                        allow_multiple_files=True
                    )

        return done

    def download_data(self, dt_fr=None, dt_to=None):
        if dt_fr is None:
            dt_fr = self.dt_fr
        if dt_to is None:
            dt_to = self.dt_to
        download_obj = self.downloader(
            dt_fr, dt_to,
            sat_id=self.sat_id,
            file_type=self.product,
            force=self.force_download
        )

        return download_obj.done

    @property
    def database(self):
        return self._database

    @database.setter
    def database(self, value):
        if isinstance(value, str):
            self._database = DatabaseModel(value)
        elif issubclass(value.__class__, DatabaseModel):
            self._database = value
        else:
            raise TypeError

    @property
    def product(self):
        return self._product

    @product.setter
    def product(self, value):
        if isinstance(value, str):
            self._product = ProductModel(value)
        elif issubclass(value.__class__, ProductModel):
            self._product = value
        else:
            raise TypeError

    @property
    def facility(self):
        return self._facility

    @facility.setter
    def facility(self, value):
        if isinstance(value, str):
            self._facility = FacilityModel(value)
        elif issubclass(value.__class__, FacilityModel):
            self._facility = value
        else:
            raise TypeError

    @property
    def instrument(self):
        return self._instrument

    @instrument.setter
    def instrument(self, value):
        if isinstance(value, str):
            self._instrument = InstrumentModel(value)
        elif issubclass(value.__class__, InstrumentModel):
            self._instrument = value
        else:
            raise TypeError
